# -*- coding: utf-8 -*-
import datetime
import os

from django.conf import settings
from m3 import date2str
from m3.actions import OperationResult
from educommon.m3 import extensions

from kinder.controllers import obs
from kinder.urls import dict_controller
from kinder.logger import error

from kinder.core.declaration.models import DeclarationPrivilege, Declaration
from kinder.core.declaration_status.models import DSS
from kinder.core.declaration.declaration_history.models import (
    DeclarationHistory
)
from kinder.core.children.models import ChildrenDelegate
from kinder.core.helpers import extract_date

from lipetsk_specifics.actions import (
    DepartmentPack, DepartmentAttributesPack,
    PrivilegeConfirmationAttributesPack, PupilTransferPrintPack,
    DeclLipetskPrintNotificationPack, ChildrenDelegatePack,
    RejectReasonPack, LipetskDelegateForDeclarationPack, PassingSmevPack)
from lipetsk_specifics.constants import (
    PRIVILEGE_CONFIRMATION_STATUSES_MAP as PCS)
from lipetsk_specifics.forms import (
    DeclarationEditWindowExtension, PrivilegeEditWindowExtension,
    DeclarationPrivilegeAddEditWindowExtension,
    PupilTransferListWindowExtension, PrintDeclNotificationWindowExtension,
    LogStatusAddWindowExtension, ChangeReasonWindow,
    DeclarationListWindowExtension)
from lipetsk_specifics.models import (
    Department,
    PrivilegeConfirmationAttributes,
    DeclarationRejectReason)

from concentrator.models import ChangeDeclaration
from concentrator.changes.rules import display_changes_map
from concentrator.change import StorageHelper
from .change import DeclarationChangeHelper
from .change import DeclarationUnitChangeHelper
from .change import LipetskDeclarationPrivilegeChangeHelper
from .change import PrivConfAttrsChangeHelper

from lipetsk_specifics.tasks import (
    ChangeInDeclarationPushTask, DecisionAboutDeclarationDirectionInQueueTask)
from lipetsk_specifics.models import RejectReason
from lipetsk_specifics.constants import (
    CHILDREN_APPLY_IN_QUEUE_TMP)
from lipetsk_specifics.changes.actions import (
    LipetskChangesPack,
    LipetskChangesDetailPack)

from .forms import PrivilegeEditWinExtender
from .constants import NOTIFICATION_MODE


def register_actions():
    dict_controller.packs.extend([
        ChildrenDelegatePack(),
        DepartmentAttributesPack(),
        DepartmentPack(),
        PrivilegeConfirmationAttributesPack(),
        PupilTransferPrintPack(),
        DeclLipetskPrintNotificationPack(),
        RejectReasonPack(),
        PassingSmevPack(),
    ])

    register_change_handlers()


def register_change_handlers():
    u"""Свои обработчики изменений."""
    StorageHelper.register_change_helper(
        'Declaration', DeclarationChangeHelper)
    StorageHelper.register_change_helper(
        'DeclarationUnit', DeclarationUnitChangeHelper)

    # расширение обработчиков изменений
    StorageHelper.register_change_helper(
        'DeclarationPrivilege', LipetskDeclarationPrivilegeChangeHelper)
    StorageHelper.register_change_helper(
        'PrivilegeConfirmationAttributes', PrivConfAttrsChangeHelper)

    # расширение функций отображения изменений
    display_changes_map.set(
        'PrivilegeConfirmationAttributes', _get_priv_conf_attr)


def _get_priv_conf_attr(change):
    """В липецке льгота всегда одна,поэтому и атрибуты подверждения будут в 1
    экземпляре, либо не будут вообще
    """
    qs = PrivilegeConfirmationAttributes.objects.filter(
        declaration_privilege__in=display_changes_map.declarationprivilege(
            change
        ).all())
    return qs[0] if qs else None


@obs.subscribe
class DeclarationListWindowListener(object):
    u"""Листенер настройки спискового окна реестра заявлений."""

    listen = [
        'kinder.core.queue_module.declaration/'
        'QueueDeclarationPack/QueueDeclaratonListWindowAction'
    ]

    def configure_list_window(self, params):
        win = params['win']
        context = params['context']

        # if settings.USE_DECL_SMEV_TAB:
        #     DeclarationListWindowExtension.add_passing_smev(win)

        return params


@obs.subscribe
class DeclarationEditWindowListener(object):
    u"""Листенер настройки окна редактирования заявления."""

    listen = [
        'kinder.core.queue_module.declaration/'
        'QueueDeclarationPack/QueueDeclarationEditAction'
    ]

    def configure_edit_window(self, params):
        win = params['win']
        context = params['context']

        DeclarationEditWindowExtension.build_extension(win)

        DeclarationEditWindowExtension.add_print_decline(win, context)
        DeclarationEditWindowExtension.history_tab_configure(win, context)
        # if settings.USE_DECL_SMEV_TAB:
        #     DeclarationEditWindowExtension.add_passing_smev_tab(win, context)

        return params


@obs.subscribe
class MakeObjectRowsActionExtension(object):
    """
    Перехватываем rows action. Добавляем статус подтверждения льготы
    """

    listen = [
        'kinder.core.declaration.declaration_privilege/'
        'DeclarationPrivilegePack/ObjectRowsAction'
    ]

    def get_rows(self, res):
        for row in res:
            declaration_privilege = self.action.parent.get_row(row['id'])

            try:
                privilegeconfirmationattributes = (
                    declaration_privilege.
                    privilegeconfirmationattributes_set.
                    latest('id'))
            except PrivilegeConfirmationAttributes.DoesNotExist:
                confirmed = False
            else:
                confirmed = privilegeconfirmationattributes.confirmed

            row['status'] = PCS[confirmed]

        return res


@obs.subscribe
class MakeObjectSaveActionExtension(object):
    """
    Перехватываем save action для возврата id записи
    """

    listen = [
        # 'kinder.core.declaration.declaration_privilege/'
        # 'DeclarationPrivilegePack/ObjectSaveAction',
        # для ЭП пришлось поменять action, но неизвестно навсегда ли
        'kinder.core.declaration.declaration_privilege/'
        'DeclarationPrivilegePack/SignatureSaveAction',
    ]

    def bind_additional_fields(self, context, instance):
        win = DeclarationPrivilegeAddEditWindowExtension
        for field_name in win.additional_privilege_attributes_fields:
            value = getattr(context, field_name)
            if value:
                setattr(instance, field_name, value)

    # Перехватываем для восстановления полей doc_* т.к. поля на другой форме
    # И для определения изменения льготы
    def save_object(self, obj):
        if obj.id:
            try:
                orig_obj = DeclarationPrivilege.objects.get(id=obj.id)
            except DeclarationPrivilege.DoesNotExist:
                pass
            else:
                obj.doc_type_id = orig_obj.doc_type_id
                obj.doc_series = orig_obj.doc_series
                obj.doc_number = orig_obj.doc_number

                try:
                    confirmation = obj.privilegeconfirmationattributes_set.latest('id')
                except PrivilegeConfirmationAttributes.DoesNotExist:
                    pass
                else:
                    if obj.privilege_id != orig_obj.privilege_id:
                        confirmation.confirmed = False
                    confirmation.save()

        return obj

    def after(self, request, context, result):
        try:
            declaration_privilege = DeclarationPrivilege.objects.get(
                declaration_id=context.declaration_id,
                privilege_id=context.privilege_id)
        except DeclarationPrivilege.DoesNotExist:
            pass
        else:
            try:
                confirmation = (
                    declaration_privilege.privilegeconfirmationattributes_set.latest('id'))
                confirmation.privilege_owner = context.privilege_owner or None
            except PrivilegeConfirmationAttributes.DoesNotExist:
                confirmation = (
                    declaration_privilege.privilegeconfirmationattributes_set.create(
                        privilege_owner=context.privilege_owner or None))

            if getattr(context, 'dismissal_date'):
                context.dismissal_date = extract_date(
                    request, 'dismissal_date')
            self.bind_additional_fields(context, confirmation)
            confirmation.delegate_id = context.delegate_id or None
            confirmation.save()

            if context.privilege_confirmation == 'true':
                result.code = u"""(
                function(){
                    win = Ext.getCmp('%s');
                    context = win.actionContextJson;
                    context.decl_privilege_id = %i;
                    context.privilege_confirmation = false;
                    context.privilege_owner = %i;
                }()
                )""" % (context.m3_window_id,
                        declaration_privilege.id,
                        int(context.privilege_owner))


@obs.subscribe
class MakeObjectDeleteActionExtension(object):
    """
    Перехватываем delete action для удаления связанных записей
    """

    listen = [
        'kinder.core.declaration.declaration_privilege/'
        'DeclarationPrivilegePack/ObjectDeleteAction'
    ]

    def before(self, request, context):
        PrivilegeConfirmationAttributes.objects.filter(
            declaration_privilege_id__in=context.decl_privilege_id).delete()


@obs.subscribe
class AddDepartment(object):
    """
    Добавляем поле "Ведомство" для льготы
    """

    listen = [
        'kinder.core.privilege/'
        'PrivilegeDictPack/ObjectEditWindowAction',
    ]

    def create_window(self, win):
        return PrivilegeEditWindowExtension.build_extension(win)

    def after(self, request, context, result):
        result.data = PrivilegeEditWindowExtension.set_value(
            result.data, self.action, context)


@obs.subscribe
class MakeDictSaveActionExtension(object):
    """
    Перехватываем экшн сохранения льготы
    Добавляем привязку льготы к ведомству
    """

    listen = [
        'kinder.core.privilege/'
        'PrivilegeDictPack/ObjectSaveAction'
    ]

    def after(self, request, context, result):
        #TODO:Тут всегда пересохранется привязка ведомства к льготе
        # , надо исправить, и реализовать отображение привязок
        priv_pack = self.action.parent

        privilege_id = getattr(context, priv_pack.id_param_name)
        department_id = getattr(context, DepartmentPack.id_param_name, None)

        if privilege_id:
            privilege = self.action.parent.get_row(privilege_id)
            privilege.department_set.clear()

            if department_id:
                try:
                    department = Department.objects.get(id=department_id)
                except Department.DoesNotExist:
                    pass
                else:
                    privilege.department_set.add(department)


@obs.subscribe
class AddPrivilegeConfirmationButton(object):
    """
    Добавляем статус подтверждения льготы
    Добавляем кнопку подтверждения льготы
    Убираем поля документа
    """

    listen = [
        'kinder.core.declaration.declaration_privilege/'
        'DeclarationPrivilegePack/ObjectEditWindowAction',
        'kinder.core.declaration.declaration_privilege/'
        'DeclarationPrivilegePack/ObjectAddWindowAction'
    ]

    def after(self, request, context, result):
        result.data = DeclarationPrivilegeAddEditWindowExtension.build_extension(
            result.data, self.action, context)


@obs.subscribe
class PrintPupilTransferExtension(object):
    """
    Кнопка печати информации о переводе ребенка
    """

    listen = [
        'kinder.core.pupil_transfer/PupilTransferPack'
        '/PupilTransferListWindowAction'
    ]

    def after(self, request, context, result):
        result.data = PupilTransferListWindowExtension.build_extension(
            result.data)


@obs.subscribe
class PrintDeclNotificationExtension(object):
    """
    Перехватываем печать уведомления для заявлений
    """

    listen = [
        'kinder.core.queue_module.declaration'
        '/QueueDeclarationPack/QueueDeclDelegateChoiceAction'
    ]

    @staticmethod
    def get_notification_mode(status_code):
        if status_code == DSS.WANT_CHANGE_DOU:
            return NOTIFICATION_MODE['WANT_CHANGE_DOU']
        elif status_code in [DSS.REGISTERED, DSS.PRIV_CONFIRMATING]:
            return NOTIFICATION_MODE['IN_QUEUE']
        elif status_code == DSS.REFUSED:
            return NOTIFICATION_MODE['REJECT']
        return None

    def after(self, request, context, result):
        try:
            decl_status_code = Declaration.objects.get(
                id=context.declaration_id
            ).status.code
        except Declaration.DoesNotExist:
            pass
        else:
            if (
                result.data and
                context.kind == result.data.PRINT_NOTIFICATION and
                self.get_notification_mode(decl_status_code)
            ):
                if getattr(context, 'printDecline', None):
                    result.data = (
                        PrintDeclNotificationWindowExtension.change_form_url(
                            result.data
                        )
                    )
                else:
                    result.data = (
                        PrintDeclNotificationWindowExtension.build_extension(
                            result.data,
                            self.get_notification_mode(decl_status_code)
                        )
                    )


@obs.subscribe
class AddStatusForNotificationExtension(object):
    """
    Добавляем кнопку об уведомления в окно заявки в статусе отказано
    """

    listen = [
        'kinder.core.queue_module.declaration'
        '/QueueDeclarationPack/QueueDeclarationEditAction'
    ]

    def additional_statuses(self, statuses):
        return statuses + (DSS.REFUSED,)


@obs.subscribe
class PupilDeclarationExtension(object):
    """
    История о том как Липецк печать заявления переопределял
    """
    listen = [
        'kinder.core.declaration/DeclarationPackPreview/DeclPrintAction',
        'kinder.core.declaration/DeclarationPack/DeclPrintAction'
    ]

    def additional_param(self, data):

        def get_priv_info(delegate):
            psas = PrivilegeConfirmationAttributes.objects.filter(
                delegate=delegate)
            if not psas:
                return dict(priv_owner='',
                            priv_owner_info='',
                            )
            psa = psas[0]
            owner_info = u''
            priv_owner_info = dict(
                document_issued_by=(
                    u'Кем выдан: ', psa.document_issued_by),
                document_date=(u'Когда выдан', psa.document_date),
                personal_number=(u'Личный номер', psa.personal_number),
                force_kind=(u'Вид войск', psa.force_kind),
                military_unit=(
                    u'Воинская часть (Подразделение)', psa.military_unit),
                dismissal_date=(u'Дата увольнения', psa.dismissal_date),
                rank=(u'Звание', psa.rank),
            )
            for key, value in priv_owner_info.iteritems():
                if value[1]:
                    owner_info += u'%s: %s\n' % (value[0], value[1])

            return dict(priv_owner=delegate.fullname,
                        priv_owner_info=owner_info,
                        )

        ext_dict = dict(
            zags_number=data['child'].zags_act_number,
            zags_date=data['child'].zags_act_date.strftime('%d.%m.%Y') if
            data['child'].zags_act_date else '',
            zags_place=data['child'].zags_act_place,
            reg_address_full=data['child'].reg_address_full,
            delegate_address=data['delegate'].address_full,
            priv_owner=get_priv_info(data['delegate'])['priv_owner'],
            priv_owner_info=get_priv_info(data['delegate'])['priv_owner_info'],
            current_date=datetime.date.today().strftime('%d.%m.%Y'),
        )
        data['variables'].update(ext_dict)
        return data

    def get_template(self, *args, **kwargs):
        template_name = 'templates/xls/declaration_info_lipetsk.xls'
        template_url = os.path.join(os.path.dirname(__file__), template_name)
        return template_url

    def get_html_template(self, *args, **kwargs):
        return os.path.join(os.path.dirname(__file__),
                            'templates/html/declaration_info_lipetsk.html')


@obs.subscribe
class AddDeclarationRejectReason(object):
    """
    Добавляем поле "Причина отказа" на форму смены статуса заявки
    """

    listen = [
        'kinder.core.declaration.declaration_status_log'
        '/LogStatusPack/ObjectAddWindowAction'
    ]

    def create_window(self, win):
        return LogStatusAddWindowExtension.build_extension(win)


@obs.subscribe
class SaveDeclarationRejectReason(object):
    """
    Перехватываем действие сохранения статуса заявления
    Создаем/обновляем или удаляем связку заявления и причины отказа
    """

    priority = 99

    listen = [
        'kinder.core.declaration.declaration_status_log'
        '/LogStatusPack/SaveAction'
    ]

    def after(self, request, context, result):
        if result.success and hasattr(context, 'reject_reason_id'):
            declaration_reject_reason, _ = (
                DeclarationRejectReason.objects.get_or_create(
                    declaration_id=context.declaration_id))
            declaration_reject_reason.reject_reason_id = (
                context.reject_reason_id)
            declaration_reject_reason.save()
        else:
            DeclarationRejectReason.objects.filter(
                declaration_id=context.declaration_id).delete()


@obs.subscribe
class AfterRejectHandler(object):
    """Отправка информации на ЕПГУ"""
    # Номера от 90 до 100 зарезервированы для перехвата зпуска комплектования
    priority = 100

    listen = [
        'kinder.core.declaration.declaration_status_log'
        '/LogStatusPack/SaveAction'
    ]

    def after(self, request, context, result):

        if result.success:

            try:
                declaration = Declaration.objects.get(
                    id=context.declaration_id)

                apply_in_queue = declaration.status.code in [
                    DSS.REGISTERED, DSS.PRIV_CONFIRMATING]
                reject_in_queue = hasattr(context, "reject_reason_id")

                if reject_in_queue:
                    decision = False
                    try:
                        reject_reason = RejectReason.objects.get(
                            id=context.reject_reason_id)
                        comment = reject_reason.name
                    except RejectReason.DoesNotExist as e:
                        error(
                            u"Ошибка при попытке отправить "
                            u"pushEvent: Причина отказа "
                            u"с id={id} не найдена".format(
                                id=e.msg))
                        comment = ""
                elif apply_in_queue:
                    comment = CHILDREN_APPLY_IN_QUEUE_TMP.format(
                        client_id=declaration.client_id)
                    decision = True

                if apply_in_queue or reject_in_queue:

                    DecisionAboutDeclarationDirectionInQueueTask().apply_async(
                        kwargs={
                            "declaration": declaration,
                            "decision": decision,
                            "profile": request.user.get_profile(),
                            "comment": comment,
                        }
                    )
            except Declaration.DoesNotExist as e:
                error(
                    u"Ошибка при попытке отправить "
                    u"pushEvent: Заявка с "
                    u"client_id={client_id} не найдена".format(
                        client_id=e.msg))


@obs.subscribe
class AfterChangeReject(object):

    listen = [
        "concentrator.changes"
        "/ChangesDetailPack/ChangesRejectAction"
    ]

    def after(self, request, context, result):

        changes = ChangeDeclaration.objects.get(
            id=context.id)

        ChangeInDeclarationPushTask().apply_async(
            kwargs={
                "declaration": changes.declaration,
                "case_number": changes.case_number,
                "decision": False})


@obs.subscribe
class AfterChangeApply(object):

    listen = [
        "concentrator.changes"
        "/ChangesDetailPack/ChangesApplyAction"
    ]

    def after(self, request, context, result):

        changes = ChangeDeclaration.objects.get(
            id=context.id)

        ChangeInDeclarationPushTask().apply_async(
            kwargs={
                "declaration": changes.declaration,
                "case_number": changes.case_number,
                "decision": True})


# подмена паков
action_pack_overrides = {
    'kinder.core.children.actions.DelegateForDeclarationPack':
        LipetskDelegateForDeclarationPack(),
    'concentrator.changes.actions.ChangesPack':
        LipetskChangesPack(),
    'concentrator.changes.actions.ChangesDetailPack':
        LipetskChangesDetailPack()}


@obs.subscribe
class SaveDeclarationEditTime(object):

    listen = [
        'kinder.core.queue_module.declaration/'
        'QueueDeclarationPack/QueueDeclarationEditAction',
        'kinder.core.queue_module.declaration/'
        'QueueDeclarationPack/DeclarationAddAction'
    ]

    def after(self, request, context, response):
        # пихаем в контекст время открытия окна Заявки
        # оно нам будет необходимо при закрытии окна
        # см. DeclarationSaveListener
        context.timestamp = date2str(
            datetime.datetime.now(), '%d.%m.%Y %H:%M:%S')


@obs.subscribe
class DeclarationSaveListener(object):
    """
    Делаем обработку после сохранения заявки
    """

    listen = [
        'kinder.core.queue_module.declaration'
        '/QueueDeclarationPack/DeclarationSaveAction'
    ]

    def after(self, request, context, response):
        def get_values(new_value=True, **kwargs):
            value_name = 'new_value' if new_value else 'old_value'

            values = u', '.join(
                history_records.filter(
                    new_value__isnull=False, **kwargs
                ).values_list(
                    value_name, flat=True
                )
            )
            return values

        if isinstance(response.data, dict) and response.data.get('success'):
            win_timestamp = datetime.datetime.strptime(
                context.timestamp, '%d.%m.%Y %H:%M:%S'
            )
            # используем время создания окна для выдергивания истории изменений
            history_records = DeclarationHistory.objects.filter(
                declaration_id=context.declaration_id,
                user_id=request.user.id,
                date__gte=win_timestamp,
            )
            changes = {
                'desired_unit': get_values(table='declaration_unit'),
                'desired_date': get_values(
                    table='declaration', field='desired_date'),
                'fullname': get_values(
                    table='children', field='fullname', new_value=False),
                'address': get_values(table='children', field='address_full'),
                'reg_address': get_values(
                    table='children', field='reg_address_full'),
                'phone': get_values(table='delegate', field='phones'),
                'email': get_values(table='delegate', field='email'),
                'privileges': get_values(table='declaration_privilege'),
            }
            # если было хоть какое-то из отслежываемых изменений
            if any(changes.values()):
                delegates = ChildrenDelegate.objects.filter(
                    children_id=context.children_id
                )
                if not delegates.exists():
                    return OperationResult.by_message(
                        u'У ребенка отсутствует законный представитель, '
                        u'необходимо внести сведения о нем'
                    )
                win = ChangeReasonWindow()
                win.set_params({
                    'declaration_id': context.declaration_id,
                    'delegates': delegates.values_list(
                        'delegate__id', 'delegate__fullname'
                    ),
                    'changes': changes,
                })
                response = OperationResult(
                    success=True,
                    code=win.get_script()
                )
                return response


@obs.subscribe
class PrivilegeEditWinLister(extensions.BaseEditWinListener):

    listen = (
        "kinder.core.privilege/PrivilegeDictPack/ObjectAddWindowAction",
        "kinder.core.privilege/PrivilegeDictPack/ObjectEditWindowAction")

    parent_model_field = 'privilege'
    ui_extender_cls = PrivilegeEditWinExtender

    def _get_id(self, context):
        return getattr(context, self.action.parent.id_param_name, None)


@obs.subscribe
class PrivilegeSaveListener(extensions.BaseSaveListener):

    listen = ("kinder.core.privilege/PrivilegeDictPack/ObjectSaveAction",)

    parent_model_field = 'privilege'
    ui_extender_cls = PrivilegeEditWinExtender

    def _declare_additional_context(self):
        return {
            'name_to_query_at_depart': {
                'type': 'unicode',
                'default': u''
            },
            'report_type': {
                'type': 'int'
            }
        }
